//
// Created by jedi on 25.06.21.
//

#include "lux.h"

#include <string.h>
#include <FreeRTOS.h>
#include <task.h>

#include <esp/spi.h>
#include <ws2812_i2s/ws2812_i2s.h>

extern "C" {
#include <sysparam.h>
}

const int signal_led_pin = 2;

const int cs0 = 15;

const int gpio4 = 4;
const int gpio5 = 5;

struct apa10xx_pixel_t {
    struct {
        unsigned int mod: 5, marker: 3;
    } __attribute__((packed)) global = {0x1F, 0x7};
    uint8_t b = 0;
    uint8_t g = 0;
    uint8_t r = 0;
};

static ws2812_pixel_t next_colour(int i) {
    ws2812_pixel_t colour = {{0, 0, 0, 0}};
    if(i == 8) {
        colour.white = 32;
    } else {
        colour.red = i & 1 ? 32 : 0;
        colour.green = i & 2 ? 32 : 0;
        colour.blue = i & 4 ? 32 : 0;
    }

    return colour;
}

static apa10xx_pixel_t next_color(int i) {
    apa10xx_pixel_t colour;
    colour.global.mod = 8;
    colour.r = i & 1 ? 32 : 0;
    colour.g = i & 2 ? 32 : 0;
    colour.b = i & 4 ? 32 : 0;

    return colour;
}

namespace fiatlux {

    namespace signal {
        void write_data(bool data) {
            gpio_write(signal_led_pin, !data);
        }
    }

    namespace relais {
        void write_data(uint8_t data[2]) {
            gpio_write(gpio4, (bool) data[0]);
            gpio_write(gpio5, (bool) data[1]);
        }
    }

    namespace spi_dimmer {
        void write_data(uint16_t data[6]) {
            for (int i = 0; i < 6; ++i) {
                int dac_val = (data[i] << 2) & 0x3FFC;

                spi_transfer_8(1, ~(0x00));
                gpio_write(cs0, 1);
                gpio_write(cs0, 0);
                spi_transfer_8(1, ~(0x01 << i));
                gpio_write(cs0, 1);
                gpio_write(cs0, 0);

                spi_transfer_16(1, dac_val);

                spi_transfer_8(1, ~(0x00));
                gpio_write(cs0, 1);
                gpio_write(cs0, 0);
                spi_transfer_8(1, ~(0x01 << i));
                gpio_write(cs0, 1);
                gpio_write(cs0, 0);
            }
        }
    }

    namespace ws28x {
        void write_data(ws2812_pixel_t *data, size_t len) {
            ws2812_i2s_init(len, PIXEL_RGBW);
            ws2812_i2s_update(data, PIXEL_RGBW);
        }
    }

    namespace apa10x {
        void write_data(apa10xx_pixel_t *data, size_t len) {
            spi_transfer_32(1, 0x00000000);
            for (size_t i = 0; i < len; i++)
                spi_transfer_32(1, *(uint32_t *) &data[i]);
            spi_transfer_32(1, 0xFFFFFFFF);
            spi_transfer_32(1, 0xFFFFFFFF);
        }
    }

    struct hal_error_t {
        constexpr hal_error_t() = default;

        hal_error_t(const char *) {}

        hal_error_t(const char *, hal_error_t *cause) {}
    };

    constexpr hal_error_t empty_error;

    enum class hal_module_t {
        SIGNAL, RELAIS, SPI_DIMMER, WS28X, APA10X
    };

    hal_error_t write_channel(uint8_t *data, size_t count, size_t stride, hal_module_t mod) {
        if(mod == hal_module_t::SIGNAL) {
            if(count != 1)
                return "unsupported value for count";
            if(stride != 1)
                return "unsupported value for stride";
            signal::write_data(data[0]);
        } else if(mod == hal_module_t::RELAIS) {
            if(count != 2)
                return "unsupported value for count";
            if(stride != 1)
                return "unsupported value for stride";
            relais::write_data(data);
        } else if(mod == hal_module_t::SPI_DIMMER) {
            if(count != 6)
                return "unsupported value for count";
            if(stride != 2)
                return "unsupported value for stride";
            spi_dimmer::write_data((uint16_t *) data);
        } else if(mod == hal_module_t::WS28X) {
            if(stride != 4)
                return "unsupported value for stride";
            ws28x::write_data((ws2812_pixel_t *) data, count);
        } else if(mod == hal_module_t::APA10X) {
            if(stride != 4)
                return "unsupported value for stride";
            apa10x::write_data((apa10xx_pixel_t *) data, count);
        } else {
            return "unsupported module";
        }
    }

    hal_error_t setup_channel(size_t count, size_t stride, hal_module_t mod) {
        /*if(mod == hal_module_t::SIGNAL) {
            if(count != 1)
                return "unsupported value for count";
            if(stride != 1)
                return "unsupported value for stride";
            signal::write_data(data[0]);
        } else if(mod == hal_module_t::RELAIS) {
            if(count != 2)
                return "unsupported value for count";
            if(stride != 1)
                return "unsupported value for stride";
            relais::write_data(data);
        } else if(mod == hal_module_t::SPI_DIMMER) {
            if(count != 6)
                return "unsupported value for count";
            if(stride != 2)
                return "unsupported value for stride";
            spi_dimmer::write_data((uint16_t *) data);
        } else if(mod == hal_module_t::WS28X) {
            if(stride != 4)
                return "unsupported value for stride";
            ws28x::write_data((ws2812_pixel_t *) data, count);
        } else if(mod == hal_module_t::APA10X) {
            if(stride != 4)
                return "unsupported value for stride";
            apa10x::write_data((apa10xx_pixel_t *) data, count);
        } else {
            return "unsupported module";
        }
        return empty_error;*/
    }

}

ws2812_pixel_t **pixels_ptr;

extern "C" void signal_led(bool state) {
    fiatlux::signal::write_data(state);
}


/* This task uses the high level GPIO API (esp_gpio.h) to blink an LED.
 *
 */
extern "C" void lux_task(void *pvParameters) {

    //gpio_enable(9, GPIO_INPUT);
    //gpio_enable(10, GPIO_INPUT);

    int32_t lux_ws2812_number = 8;
    auto ret = sysparam_get_int32("lux_ws2812_number", &lux_ws2812_number);
    if(ret != SYSPARAM_OK)
        lux_ws2812_number = 8;

    int32_t lux_apa10xx_number = 8;
    ret = sysparam_get_int32("lux_apa10xx_number", &lux_apa10xx_number);
    if(ret != SYSPARAM_OK)
        lux_apa10xx_number = 8;

    ws2812_pixel_t pixels[lux_ws2812_number];
    ws2812_i2s_init(lux_ws2812_number, PIXEL_RGBW);
    memset(pixels, 0, sizeof(ws2812_pixel_t) * lux_ws2812_number);

    apa10xx_pixel_t leds[lux_apa10xx_number];

    //lux_apa102c_number

    gpio_enable(signal_led_pin, GPIO_OUTPUT);
    //gpio_enable(cs0, GPIO_OUTPUT);
    gpio_enable(gpio4, GPIO_OUTPUT);
    gpio_enable(gpio5, GPIO_OUTPUT);
    //spi_init(1, SPI_MODE0, SPI_FREQ_DIV_1M, 1, SPI_BIG_ENDIAN, 1); //for spi dimmer

    spi_init(1, SPI_MODE0, SPI_FREQ_DIV_1M, 1, SPI_LITTLE_ENDIAN, false); //for apa10xx
    while (1) {
        /*for (int j = 0; j < 64; j++) {
            for (int i = 0; i < 8; i++)
                spi_dac(i, 64 * j);
            //printf("> %d\n", 64*j);
            vTaskDelay(100 / portTICK_PERIOD_MS);
        }*/
        /*gpio_write(gpio4, 1);
        vTaskDelay(200 / portTICK_PERIOD_MS);
        gpio_write(gpio4, 0);
        for (int i = 0; i < 8; i++)
            spi_dac(i, 0);

        gpio_write(gpio5, 1);
        vTaskDelay(200 / portTICK_PERIOD_MS);
        gpio_write(gpio5, 0);*/
        signal_led(0);
        for (int c = 8; c >= 0; c--) {

            for (auto &pixel: pixels) {
                pixel = next_colour(c);
            }
            for (auto &led: leds) {
                led = next_color(c);
            }
            ws2812_i2s_update(pixels, PIXEL_RGBW);
            write_leds(leds, lux_apa10xx_number);
            vTaskDelay(200 / portTICK_PERIOD_MS);
        }
        gpio_write(gpio4, 1);
        vTaskDelay(200 / portTICK_PERIOD_MS);
        gpio_write(gpio4, 0);
        gpio_write(gpio5, 1);
        vTaskDelay(200 / portTICK_PERIOD_MS);
        gpio_write(gpio5, 0);
        signal_led(1);
        vTaskDelay(200 / portTICK_PERIOD_MS);

    }
}